/***************************************************************************

    disco.h - Disco header file

    Date: 04.28.2003
    Ver : 1.0

    Disco: The passive ip discovery and fingerprinting tool

    Copyright (c) 2003 by Preston Wood
    All rights reserved.

    Author(s): Preston Wood  <p@altmode.com>

/***************************************************************************

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
  THE AUTHOR, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM,
  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************/

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <pcap.h>
#include <signal.h>

#define SNAPLEN         74
#define PROMISC         1
#define TIMEOUT         500
#define FILTER          "ip"
#define HASH_TABLE_SIZE 19999
#define MAXFP           5000

#define TCPOPT_EOL              0
//#define TCPOPT_NOP              0x01
#define TCPOPT_MAXSEG           2
#define TCPOPT_WSCALE           3
#define TCPOPT_SACKOK           4
#define TCPOPT_TIMESTAMP        8

#define EXTRACT_BIT(p) \
        ((u_short)*((u_char *)(p) + 0))


#define EXTRACT_16BITS(p) \
        ((u_short)*((u_char *)(p) + 0) << 8 | \
        (u_short)*((u_char *)(p) + 1))

#define EXTRACT_24BITS(p) \
        ((u_short)*((u_char *)(p) + 0) << 8 | \
        (u_short)*((u_char *)(p) + 1) << 8 | \
        (u_short)*((u_char *)(p) + 2))
        
#define IP_DF   0x4000  /* dont fragment flag */
#define IP_MF   0x2000  /* more fragments flag */

#define TCP_FIN  0x01
#define TCP_SYN  0x02
#define TCP_RST  0x04
#define TCP_PUSH 0x08
#define TCP_ACK  0x10
#define TCP_URG  0x20

#define targsize 255


#ifdef _SUN_
typedef   unsigned char u_int8_t;
typedef   unsigned int u_int32_t;
#endif

struct table_entry
{
    uint32_t uip;
    struct table_entry *next;
};

struct fingerprints
{
    int       win;
    int       ttl;
    u_int32_t mss;
    u_int8_t  df;
    int       wscale;
    u_int16_t sackok;
    u_int8_t  nop;
    u_int16_t psize;
    char      *os;
};

    
char *iprintf(u_char *);
int packet_to_check(uint32_t, struct table_entry **);
char fingerprint_packet(u_char *);
int load_fingerprints();
int packet_dup_check(uint32_t, struct table_entry **, uint32_t);
int packet_add_entry(uint32_t, struct table_entry **, uint32_t);
uint32_t packet_hash(uint32_t);
void ht_init_table(struct table_entry **);
void cleanup(int);
int catch_sig(int, void(*)());
void usage(char *);

/* EOF */
