/***************************************************************************

    main.c - Disco main c file

    Date: 04.28.2003
    Ver : 1.0

    Disco: The passive ip discovery and fingerprinting tool

    Copyright (c) 2003 by Preston Wood
    All rights reserved.

    Author(s): Preston Wood  <p@altmode.com>
  
/***************************************************************************
                                                                          
  This program is free software; you can redistribute it and/or modify  
  it under the terms of the GNU General Public License as published by  
  the Free Software Foundation; either version 2 of the License, or     
  (at your option) any later version.                                   
                                                                          
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
  THE AUTHOR, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM,
  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************/
 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "./disco.h"
/* disco.c */

int loop = 1;
u_long ip = 0;
int allip = 0;
int fingerprint = 0;
int totfp = 0;
int totpacket_fp = 0;
int unique_ip=0;
int found=0;
int fp_cand=0;
char found_os[255];
int synack=0;
uint32_t srcip;
struct fingerprints loadedfp[MAXFP];
struct fingerprints packet_fp;
struct table_entry *fhash_table[HASH_TABLE_SIZE];
struct table_entry *ihash_table[HASH_TABLE_SIZE];
struct iphdr *ip_h;
struct tcphdr *tcp_h;


int main(int argc, char **argv)
{
    int c=0;         
    int totfp=0;
    pcap_t *p;
    char *device, *rule = NULL;
    char *foundip;
    char *logfile = NULL;
    u_char *packet;
    int print_ip=1;
    int iphlen=0;
    int only_syn=0;
    struct pcap_pkthdr h;
    struct pcap_stat ps;
    struct in_addr *paaddr;
    char errbuf[PCAP_ERRBUF_SIZE];
    struct bpf_program filter_code;
    bpf_u_int32 local_net, netmask;
    FILE *outfile;

    while ((c = getopt(argc, argv, "fISi:r:o:uh")) != -1)
    {
            switch (c)
            {
                     case 'f':
                        totfp = load_fingerprints();
                        fingerprint = 1;
                        break;
                     case 'N':
                        print_ip = 0;
                        break;
                     case 'S':
                        only_syn = 1;
                        break;
                     case 'i':
                        device = optarg;
                        break;
                     case 'r':
                        rule = optarg;
                        break;
                     case 'u':
                        unique_ip = 1;
                        break;
                     case 'o':
                        logfile = optarg;
                        break;
                     case 'h':
                        usage("Options");
                        break;
                     default:
                        usage("Bad Option");
            }
            if ( rule == NULL )
            {
                rule = "ip";
            }
    }

    if ( c==0 )
    {
        usage("Need Option");
    }
    
    printf("Disco v0.1\n");

    /* If no device is specified error out */

    if (device == NULL)
    {
        device = pcap_lookupdev(errbuf);
        if (device == NULL)
        {
            fprintf(stderr, "pcap_lookupdev() failed: %s\n", errbuf);
        }
    }

    /* Open the ethernet interface for packet capture */

    p = pcap_open_live(device, SNAPLEN, PROMISC, TIMEOUT, errbuf);
    if (p == NULL)
    {
        fprintf(stderr, "pcap_open_live() failed: %s\n", errbuf);
        exit(EXIT_FAILURE);
    }

    /* Set BPF filter only looking at IP packets */

    if (pcap_lookupnet(device, &local_net, &netmask, errbuf) == -1)
    {
        fprintf(stderr, "pcap_lookupnet() failed: %s\n", errbuf);
        pcap_close(p);
        exit(EXIT_FAILURE);
    }

    /* Compile rule expression into a program */

    if (pcap_compile(p, &filter_code, rule, 1, netmask) == -1)
    {
        fprintf(stderr, "pcap_compile() failed: %s\n", pcap_geterr(p));
        pcap_close(p);
        exit(EXIT_FAILURE);
    }

    /* Load the compiled filter in the packet capture device */

    if (pcap_setfilter(p, &filter_code) == -1)
    {
        fprintf(stderr, "pcap_setfilter() failed: %s\n", pcap_geterr(p));
        pcap_close(p);
        exit(EXIT_FAILURE);
    }

    /* Check the link to make sure it is ethernet */

    if (pcap_datalink(p) != DLT_EN10MB)
    {
        fprintf(stderr, "Disco only works on ethernet networks\n");
        pcap_close(p);
        exit(EXIT_FAILURE);
    }

    /* Gather captured packet stats before we exit */

    if (catch_sig(SIGINT, cleanup) == -1)
    {
        fprintf(stderr, "Can't catch the signal\n");
        pcap_close(p);
        exit(EXIT_FAILURE);
    }

    ht_init_table(ihash_table);
    outfile = fopen(logfile,"w");
    
    /* Start looping through packets ctrl-c to exit */

    for ( ht_init_table(fhash_table); loop;)
    {
        packet = (u_char *)pcap_next(p, &h);
        if (packet == NULL)
        {
            continue;
        }

        /* Parse IP Hdr to check */
        ip_h = (struct iphdr *) (packet + 14);
        iphlen = ((ip_h->ihl*4));

        srcip = (ntohl(ip_h->saddr)); //Convert src network ip to host order for easier handling

        if (only_syn == 0)
        {
            if (packet_to_check(srcip, ihash_table))  //Pass src to check for dupes if none print output
            {
                if ( logfile != NULL )
                {
                    fprintf(outfile, "%s,\n", iprintf(packet + 26));
                }
                if ( print_ip == 1 )
                {
                    printf("%s\n", iprintf(packet + 26));
                }
            }
        }
          
        if ( fingerprint==1 && unique_ip==1 ) // If fingerprint flag and unique flag is set
        {
          if (ip_h->protocol==6)
          {
            switch (ip_h->ihl)
            {
              case 5:     //IP Header length of 20 - no options
                tcp_h = (struct tcphdr *) (packet + (iphlen+14));
                //tcp_opt = (struct tcphdr *) (packet + iphlen);
                break;
              default:    //Parse through IP options
                tcp_h = (struct tcphdr *)(packet+14+(iphlen<<2));
                break;
            }
          
            if (tcp_h->syn==1 && tcp_h->ack==0)
            {
                if (packet_to_check(srcip, fhash_table))
                {  
                    fingerprint_packet(packet);

                    if ( found==1 ) //If fingerprint match is found for packet output IP and OS
                    {
                        foundip = iprintf(packet + 26);
                        if ( logfile != NULL )
                        {
                            fprintf(outfile, "%s, %s\n", foundip, found_os);
                        }
                        if ( print_ip == 1 )
                        {
                            printf("%s: %s\n", foundip, found_os); 
                        }
                        strcpy (found_os, "");
                        totpacket_fp++; //Increment total IP counter
                    }              
                    if ( found==0 )  //If no fingerprint match is found for packet output IP and fingerprint
                    {
                        foundip = iprintf(packet + 26);
                        if ( logfile != NULL )
                        {
                            fprintf(outfile, "%s,%d:%d:%d:%d:%d:%d:%d:%d\n",foundip,packet_fp.win, packet_fp.ttl, packet_fp.mss, packet_fp.df, packet_fp.wscale, packet_fp.sackok, packet_fp.nop, packet_fp.psize);
                        }
                        if ( print_ip == 1 )
                        {
                            printf("%s: %d:%d:%d:%d:%d:%d:%d:%d syn\n",foundip,packet_fp.win, packet_fp.ttl, packet_fp.mss, packet_fp.df, packet_fp.wscale, packet_fp.sackok, packet_fp.nop, packet_fp.psize);
                        }                        
                        totpacket_fp++;
                    }

                    /* Reset globals for next packet to check */
                    
                    fp_cand = 0;
                    found = 0;
                    synack = 0;
                }
            }
          }
        }
        else if ( unique_ip==0 && fingerprint==1 ) //If fingerprint flag is set and unique flag is not
        {
            if (ip_h->protocol==6)
            {
                switch (ip_h->ihl)
                {
                    case 5:     //IP Header length of 20 - no options
                      tcp_h = (struct tcphdr *) (packet + (iphlen+14));
                      //tcp_opt = (struct tcphdr *) (packet + iphlen);
                      break;
                    default:    //Parse through IP options
                      tcp_h = (struct tcphdr *)(packet+14+(iphlen<<2));
                      break;
                }
                if (tcp_h->syn==1 && tcp_h->ack==0)
                {
                    fingerprint_packet(packet);
                    if ( found==1 )
                    {
                        foundip = iprintf(packet + 26);
                        if ( logfile != NULL )
                        {
                            fprintf(outfile, "%s, %s\n",foundip, found_os);
                        }
                        if ( print_ip == 1 )
                        {
                            printf("%s: %s\n", foundip, found_os);
                        }
                        strcpy (found_os, "");
                        totpacket_fp++;
                    }
                    if ( found==0 )
                    {
                        foundip = iprintf(packet + 26);
                        if ( logfile != NULL )
                        {
                            fprintf(outfile, "%s,%d:%d:%d:%d:%d:%d:%d:%d\n",foundip,packet_fp.win, packet_fp.ttl, packet_fp.mss, packet_fp.df, packet_fp.wscale, packet_fp.sackok, packet_fp.nop, packet_fp.psize);
                        }
                        if ( print_ip == 1 )
                        {
                            printf("%s: %d:%d:%d:%d:%d:%d:%d:%d syn\n",foundip,packet_fp.win, packet_fp.ttl, packet_fp.mss, packet_fp.df, packet_fp.wscale, packet_fp.sackok, packet_fp.nop, packet_fp.psize);
                        }
                        totpacket_fp++;
                    }

                    /* Reset globals for next packet to check */

                    fp_cand = 0;
                    found = 0;
                    synack = 0;
                }
            }           
        }
          
    }

    /* crtl-c exit */

    if (pcap_stats(p, &ps) == -1)
    {
        fprintf(stderr, "pcap_stats() failed: %s\n", pcap_geterr(p));
    }
    else
    {
        /* If user exits display some packet statistics */
      
        printf("\nPackets received by Disco:\t%6d\n"
                 "Packets dropped by Disco:\t%6d\n"
                 "Unique IP addresses received by Disco:\t%6ld\n",
                 ps.ps_recv, ps.ps_drop, ip);
        printf("Total IP fingerprinted by Disco: \t%6d\n", totpacket_fp);
    }
    pcap_close(p);
    return (EXIT_SUCCESS);
}


/* Routine to print hex ip in decimal */

char * iprintf(u_char *address)
{
    static char ip_addr[17];

    sprintf(ip_addr, "%3d.%3d.%3d.%3d", (address[0]&255), (address[1]&255), (address[2]&255), (address[3]&255));
    return (ip_addr);
}

/* Check packet source ip for uniqueness */

int packet_to_check(uint32_t src_ip, struct table_entry **hash_table)
{
    uint32_t n;

    n = packet_hash(src_ip);  //Function to hash src ip from packet

    allip++;  //Increment packet statistic counter
    
    if (hash_table[n]) //If entry exists in hash table
    {
        if (!packet_dup_check(src_ip, hash_table, n)) //Check hash table for src ip
        {
            if (packet_add_entry(src_ip, hash_table, n)) //If no dupes are found add src ip to hash table
            {
                ip++;
                return(1);
            }
        }
        else  // If src ip dupe is found exit
        {
            return(0);
        }
    }
    else
    {
        if (packet_add_entry(src_ip, hash_table, n))
        {
            ip++;
            return(1);
        }
    }
    return (0);
}

/* Hash source ip into unique number */

uint32_t packet_hash(uint32_t src_ip)
{
    return (src_ip %= HASH_TABLE_SIZE); //Return remainder (Modulus) from src ip divided by hash table size     
}

/* Check source ip for dups in the hash table */

int packet_dup_check(uint32_t src_ip, struct table_entry **hash_table, uint32_t loc)
{
    struct table_entry *p;
    
    for (p = hash_table[loc]; p; p = p->next)
    {
        if (p->uip == src_ip)
        {
            return (1);
        }
    }
    return (0);
}

/* If packet isn't a dup add to hash table */

int packet_add_entry(uint32_t src_ip, struct table_entry **hash_table, uint32_t loc) //Add src ip to hash table
{
    struct table_entry *p;
    
    if (hash_table[loc] == NULL)
    {
        hash_table[loc] = malloc(sizeof(struct table_entry));
        if (hash_table[loc] == NULL)
        {
            return(0);
        }
        hash_table[loc]->uip = src_ip;
        hash_table[loc]->next = NULL;
        return (1);
    }
    else
    {
        for (p = hash_table[loc]; p->next; p = p->next);
        p->next = malloc(sizeof(struct table_entry));
        if (p->next == NULL)
        {
            return (0);
        }

        p = p->next;

        p->uip = src_ip;
        p->next = NULL;
    }
    return (1);
}

/* Routine to fingerprint the SYN packet */

char fingerprint_packet(u_char *packet)
{
    int i=0;
    int count=0;
    int opt=0;
    int iphlen=0;
    int optlen=0;
    int ttlcount;
    int wscale_exists = 0;
    struct iphdr *ip_h;
    struct tcphdr *tcp_h;
    struct ethhdr *ether;

    void* opt_ptr;

    char src_ip[20], dest_ip[20];

    ip_h = (struct iphdr *) (packet + 14);

    iphlen = ((ip_h->ihl*4));

       switch (ip_h->ihl)
        {
            case 5:     //IP Header length of 20 - no options
              tcp_h = (struct tcphdr *) (packet + (iphlen+14));
              //tcp_opt = (struct tcphdr *) (packet + iphlen);
              break;
            default:    //Parse through IP options
              tcp_h = (struct tcphdr *)(packet+14+(iphlen<<2));
              break;
        }
                fp_cand = 1;
                packet_fp.win = ntohs(tcp_h->window);
                packet_fp.ttl = ip_h->ttl;
                packet_fp.sackok = 0;
                packet_fp.nop = 0;
                packet_fp.psize = ntohs(ip_h->tot_len);
                if (ip_h->frag_off==64)
                {
                  packet_fp.df = 1;
                }
                else
                {
                  packet_fp.df = 0;
                }
                opt_ptr=(void*)tcp_h+sizeof(struct tcphdr);
                optlen = (ntohs(ip_h->tot_len))-(sizeof(struct iphdr))-(sizeof(struct tcphdr));
                while ( i < optlen )
                {
                  opt=(int)(*(u_char*)(opt_ptr+i));
                  switch(opt)
                  {
                    case TCPOPT_MAXSEG:
                    {
                      packet_fp.mss = EXTRACT_16BITS(opt_ptr+2);
                      i+=3;
                      break;
                    }
                    case TCPOPT_SACKOK:
                    {
                      packet_fp.sackok = 1;
                      i+=1;
                      break;
                    }
                    case TCPOPT_NOP:
                    {
                      packet_fp.nop = 1;
                      break;
                    }
                    case TCPOPT_WSCALE:
                    {
                      packet_fp.wscale = (u_short)*((u_char *)(opt_ptr+i+2));
                      i+=2;
                      wscale_exists = 1;
                      break;
                    } 
                }
                i++;
                }
                if ( wscale_exists==0 )
                {
                  packet_fp.wscale = -1;
                }
                while ( count < totfp  && found==0 )
                {
                  if (loadedfp[count].win == packet_fp.win &&
                      loadedfp[count].mss == packet_fp.mss &&
                      loadedfp[count].df == packet_fp.df &&
                      loadedfp[count].wscale == packet_fp.wscale &&
                      loadedfp[count].sackok == packet_fp.sackok &&
                      loadedfp[count].nop == packet_fp.nop &&
                      loadedfp[count].psize == packet_fp.psize)
                  {
                      for ( ttlcount=0;ttlcount<64;ttlcount++ )
                      {
                          if (loadedfp[count].ttl == ((packet_fp.ttl)+ttlcount))
                          {
                              strcpy (found_os, loadedfp[count].os);
                              found = 1;
                              count = totfp;
                          }
                          if ( found==1 )
                          {
                              break
                              ;
                          }
                      }
                  }
                  count++;
                }
                count=0;
                if ( tcp_h->ack == 1 )
                {
                  synack=1;
                }
}

void cleanup(int signo)
{
    loop = 0;
    printf("Interrupt signal caught...\n");
}

void ht_init_table(struct table_entry **hash_table)
{
    int c;

    for (c = 0; c < HASH_TABLE_SIZE; c++)
    {
        hash_table[c] = NULL;
    }
}

int catch_sig(int signo, void (*handler)())
{
    struct sigaction action;

    action.sa_handler = handler;
    sigemptyset(&action.sa_mask);
    action.sa_flags = 0;

    if (sigaction(signo, &action, NULL) == -1)
    {
        return (-1);
    }
    else
    {
        return (1);
    }
}

int load_fingerprints()
{
    char *fp;
    char *line;
    char infile[200];
    FILE *fp_file;
	int len=0;

    fp_file=fopen("disco.fp", "r");
    
    if (!fp_file)
    {
        fprintf(stderr, "No fingerprint file found\n");
        return;
    }

    while (fgets(infile, 200, fp_file) != NULL && infile[0] != '\n')
    {
        line = strtok ( infile, "\n");
        
        fp = strtok ( line, ":");
        loadedfp[totfp].win = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].ttl = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].mss = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].df = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].wscale = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].sackok = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].nop = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].psize = atoi(fp);
        fp = strtok ( NULL, ":");
        loadedfp[totfp].os =  (char *) malloc(strlen (fp)+1);
        strcpy (loadedfp[totfp].os, fp);
        totfp++;
    }
    fclose(fp_file);
    return(totfp);
}

void usage(char *errmsg)
{
    fprintf(stderr,"Disco: %s\n", errmsg);
    fprintf(stderr,"\nUsage: disco  [ -i device ] [ -N don't print IP ] [ -f fingerprint ] [ -S SYN only ] [ -o file ]\n"
                   "                      [ -r filter rule ] [ -u unique IP ] [ -g enable gherkin ] \n");
    fprintf(stderr," -i device          capture packets from device\n");
    fprintf(stderr," -N don't print IP  Do not print IP info to STDOUT\n");
    fprintf(stderr," -f fingerprint     fingerprint syn packets\n");
    fprintf(stderr," -S SYN only        only watch for SYN packets\n");
    fprintf(stderr," -o file            write output to file\n");
    fprintf(stderr," -r filter rule     tcpdump filter rules\n");
    fprintf(stderr," -u unique ip       unique ip only applies to fingerprint option\n");
    fprintf(stderr," -h help            display options\n");
    fprintf(stderr," -g enable gherkin  output to gherkin database - not implemented yet\n");
    exit(1);
}
